@@ -69,6 +69,10 @@ enum
 
 
 
+typedef struct _SchedulerIdleInfo SchedulerIdleInfo;
+
+
+
 static void tumbler_service_constructed        (GObject            *object);
 static void tumbler_service_finalize           (GObject            *object);
 static void tumbler_service_get_property       (GObject            *object,
@@ -102,6 +106,7 @@ static void tumbler_service_scheduler_started  (TumblerScheduler   *scheduler,
 static void tumbler_service_pre_unmount        (TumblerService     *service,
                                                 GMount             *mount,
                                                 GVolumeMonitor     *monitor);
+static void scheduler_idle_info_free           (SchedulerIdleInfo  *info);
 
 
 
@@ -122,6 +127,17 @@ struct _TumblerService
   GVolumeMonitor   *volume_monitor;
 };
 
+struct _SchedulerIdleInfo
+{
+  TumblerScheduler *scheduler;
+  TumblerService   *service;
+  GStrv             uris;
+  gchar            *message;
+  gchar            *origin;
+  guint             handle;
+  gint              error_code;
+};
+
 
 
 static guint tumbler_service_signals[LAST_SIGNAL];
@@ -358,58 +374,126 @@ tumbler_service_set_property (GObject      *object,
 
 
 
-static void
-tumbler_service_scheduler_error (TumblerScheduler   *scheduler,
-                                 guint               handle,
-                                 const gchar *const *failed_uris,
-                                 gint                error_code,
-                                 const gchar        *message_s,
-                                 const gchar        *origin,
-                                 TumblerService     *service)
+static gboolean
+tumbler_service_error_idle (gpointer user_data)
 {
-  DBusMessageIter iter;
-  DBusMessageIter strv_iter;
-  DBusMessage    *message;
-  guint           n;
+  SchedulerIdleInfo *info = user_data;
+  DBusMessageIter    iter;
+  DBusMessageIter    strv_iter;
+  DBusMessage       *message;
+  guint              n;
 
-  g_return_if_fail (TUMBLER_IS_SCHEDULER (scheduler));
-  g_return_if_fail (failed_uris != NULL && failed_uris[0] != NULL && *failed_uris[0] != '\0');
-  g_return_if_fail (message_s != NULL && *message_s != '\0');
-  g_return_if_fail (origin != NULL && *origin != '\0');
-  g_return_if_fail (TUMBLER_IS_SERVICE (service));
+  g_return_if_fail (info != NULL);
+  g_return_if_fail (TUMBLER_IS_SCHEDULER (info->scheduler));
+  g_return_if_fail (info->uris != NULL && info->uris[0] != NULL && *info->uris[0] != '\0');
+  g_return_if_fail (info->message != NULL && *info->message != '\0');
+  g_return_if_fail (info->origin != NULL && *info->origin != '\0');
+  g_return_if_fail (TUMBLER_IS_SERVICE (info->service));
 
   /* create a D-Bus message for the error signal */
   message = dbus_message_new_signal (THUMBNAILER_PATH, THUMBNAILER_IFACE, "Error");
 
   /* define the destination (the thumbnailer client) if possible */
-  if (origin)
-    dbus_message_set_destination (message, origin);
+  if (info->origin)
+    dbus_message_set_destination (message, info->origin);
 
   /* append the request handle */
   dbus_message_iter_init_append (message, &iter);
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &handle);
+  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &info->handle);
 
   /* start the URI string array */
   dbus_message_iter_open_container (&iter, DBUS_TYPE_ARRAY, 
                                     DBUS_TYPE_STRING_AS_STRING, &strv_iter);
 
   /* insert all failed URIs into the array */
-  for (n = 0; failed_uris[n] != NULL; n++)
-    dbus_message_iter_append_basic (&strv_iter, DBUS_TYPE_STRING, &failed_uris[n]);
+  for (n = 0; info->uris[n] != NULL; n++)
+    dbus_message_iter_append_basic (&strv_iter, DBUS_TYPE_STRING, &info->uris[n]);
 
   /* finish the URI string array */
   dbus_message_iter_close_container (&iter, &strv_iter);
 
   /* append the error code and error message */
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &error_code);
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &message_s);
+  dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &info->error_code);
+  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &info->message);
 
   /* send the signal message over D-Bus */
-  dbus_connection_send (dbus_g_connection_get_connection (service->connection), 
+  dbus_connection_send (dbus_g_connection_get_connection (info->service->connection), 
                         message, NULL);
 
   /* free the allocated D-Bus message */
   dbus_message_unref (message);
+
+  scheduler_idle_info_free (info);
+
+  return FALSE;
+}
+
+
+
+static void
+tumbler_service_scheduler_error (TumblerScheduler   *scheduler,
+                                 guint               handle,
+                                 const gchar *const *failed_uris,
+                                 gint                error_code,
+                                 const gchar        *message,
+                                 const gchar        *origin,
+                                 TumblerService     *service)
+{
+  SchedulerIdleInfo *info;
+
+  g_return_if_fail (TUMBLER_IS_SCHEDULER (scheduler));
+  g_return_if_fail (failed_uris != NULL);
+  g_return_if_fail (message != NULL && *message != '\0');
+  g_return_if_fail (origin != NULL && *origin != '\0');
+  g_return_if_fail (TUMBLER_IS_SERVICE (service));
+  
+  info = g_slice_new0 (SchedulerIdleInfo);
+
+  info->scheduler = g_object_ref (scheduler);
+  info->handle = handle;
+  info->uris = g_strdupv ((GStrv) failed_uris);
+  info->error_code = error_code;
+  info->message = g_strdup (message);
+  info->origin = g_strdup (origin);
+  info->service = g_object_ref (service);
+
+  g_idle_add (tumbler_service_error_idle, info);
+}
+
+
+
+static gboolean
+tumbler_service_finished_idle (gpointer user_data)
+{
+  SchedulerIdleInfo *info = user_data;
+  DBusMessageIter    iter;
+  DBusMessage       *message;
+
+  g_return_val_if_fail (info != NULL, FALSE);
+  g_return_val_if_fail (info->origin != NULL && *info->origin != '\0', FALSE);
+  g_return_val_if_fail (TUMBLER_IS_SERVICE (info->service), FALSE);
+
+  /* create a D-Bus message for the finished signal */
+  message = dbus_message_new_signal (THUMBNAILER_PATH, THUMBNAILER_IFACE, "Finished");
+
+  /* define the destination (the thumbnailer client) if possible */
+  if (info->origin)
+    dbus_message_set_destination (message, info->origin);
+
+  /* append the request handle */
+  dbus_message_iter_init_append (message, &iter);
+  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &info->handle);
+
+  /* send the signal message over D-Bus */
+  dbus_connection_send (dbus_g_connection_get_connection (info->service->connection), 
+                        message, NULL);
+
+  /* free the allocated D-Bus message */
+  dbus_message_unref (message);
+
+  scheduler_idle_info_free (info);
+
+  return FALSE;
 }
 
 
@@ -420,26 +504,72 @@ tumbler_service_scheduler_finished (TumblerScheduler *scheduler,
                                     const gchar      *origin,
                                     TumblerService   *service)
 {
-  DBusMessageIter iter;
-  DBusMessage    *message;
+  SchedulerIdleInfo *info;
 
-  /* create a D-Bus message for the finished signal */
-  message = dbus_message_new_signal (THUMBNAILER_PATH, THUMBNAILER_IFACE, "Finished");
+  g_return_if_fail (TUMBLER_IS_SCHEDULER (scheduler));
+  g_return_if_fail (origin != NULL && *origin != '\0');
+  g_return_if_fail (TUMBLER_IS_SERVICE (service));
+  
+  info = g_slice_new0 (SchedulerIdleInfo);
+
+  info->scheduler = g_object_ref (scheduler);
+  info->handle = handle;
+  info->origin = g_strdup (origin);
+  info->service = g_object_ref (service);
+
+  g_idle_add (tumbler_service_finished_idle, info);
+}
+
+
+
+static gboolean 
+tumbler_service_ready_idle (gpointer user_data)
+{
+  SchedulerIdleInfo *info = user_data;
+  DBusMessageIter    iter;
+  DBusMessageIter    strv_iter;
+  DBusMessage       *message;
+  guint              n;
+
+  g_return_if_fail (info != NULL);
+  g_return_if_fail (TUMBLER_IS_SCHEDULER (info->scheduler));
+  g_return_if_fail (info->uris != NULL && info->uris[0] != NULL && *info->uris[0] != '\0');
+  g_return_if_fail (info->origin != NULL && *info->origin != '\0');
+  g_return_if_fail (TUMBLER_IS_SERVICE (info->service));
+
+  /* create a D-Bus message for the ready signal */
+  message = dbus_message_new_signal (THUMBNAILER_PATH, THUMBNAILER_IFACE, "Ready");
 
   /* define the destination (the thumbnailer client) if possible */
-  if (origin)
-    dbus_message_set_destination (message, origin);
+  if (info->origin)
+    dbus_message_set_destination (message, info->origin);
+
+  dbus_message_iter_init_append (message, &iter);
 
   /* append the request handle */
-  dbus_message_iter_init_append (message, &iter);
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &handle);
+  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &info->handle);
+
+  /* start the URI string array */
+  dbus_message_iter_open_container (&iter, DBUS_TYPE_ARRAY,
+                                    DBUS_TYPE_STRING_AS_STRING, &strv_iter);
+
+  /* insert all URIs into the array for which we have thumbnails now */
+  for (n = 0; info->uris[n] != NULL; n++)
+    dbus_message_iter_append_basic (&strv_iter, DBUS_TYPE_STRING, &info->uris[n]);
+
+  /* finish the URI string array */
+  dbus_message_iter_close_container (&iter, &strv_iter);
 
   /* send the signal message over D-Bus */
-  dbus_connection_send (dbus_g_connection_get_connection (service->connection), 
+  dbus_connection_send (dbus_g_connection_get_connection (info->service->connection), 
                         message, NULL);
 
   /* free the allocated D-Bus message */
   dbus_message_unref (message);
+
+  scheduler_idle_info_free (info);
+
+  return FALSE;
 }
 
 
@@ -451,41 +581,59 @@ tumbler_service_scheduler_ready (TumblerScheduler *scheduler,
                                  const gchar      *origin,
                                  TumblerService   *service)
 {
-  DBusMessageIter iter;
-  DBusMessageIter strv_iter;
-  DBusMessage    *message;
-  guint           n;
+  SchedulerIdleInfo *info;
+  
+  g_return_if_fail (TUMBLER_IS_SCHEDULER (scheduler));
+  g_return_if_fail (origin != NULL && *origin != '\0');
+  g_return_if_fail (uris != NULL && uris[0] != NULL && *uris[0] != '\0');
+  g_return_if_fail (TUMBLER_IS_SERVICE (service));
+  
+  info = g_slice_new0 (SchedulerIdleInfo);
 
-  /* create a D-Bus message for the ready signal */
-  message = dbus_message_new_signal (THUMBNAILER_PATH, THUMBNAILER_IFACE, "Ready");
+  info->scheduler = g_object_ref (scheduler);
+  info->handle = handle;
+  info->uris = g_strdupv ((GStrv) uris);
+  info->origin = g_strdup (origin);
+  info->service = g_object_ref (service);
+
+  g_idle_add (tumbler_service_ready_idle, info);
+}
+
+
+
+static gboolean
+tumbler_service_started_idle (gpointer user_data)
+{
+  SchedulerIdleInfo *info = user_data;
+  DBusMessageIter   iter;
+  DBusMessage      *message;
+
+  g_return_if_fail (info != NULL);
+  g_return_if_fail (TUMBLER_IS_SCHEDULER (info->scheduler));
+  g_return_if_fail (info->origin != NULL && *info->origin != '\0');
+  g_return_if_fail (TUMBLER_IS_SERVICE (info->service));
+
+  /* create a D-Bus message for the started signal */
+  message = dbus_message_new_signal (THUMBNAILER_PATH, THUMBNAILER_IFACE, "Started");
 
   /* define the destination (the thumbnailer client) if possible */
-  if (origin)
-    dbus_message_set_destination (message, origin);
-
-  dbus_message_iter_init_append (message, &iter);
+  if (info->origin)
+    dbus_message_set_destination (message, info->origin);
 
   /* append the request handle */
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &handle);
-
-  /* start the URI string array */
-  dbus_message_iter_open_container (&iter, DBUS_TYPE_ARRAY,
-                                    DBUS_TYPE_STRING_AS_STRING, &strv_iter);
-
-  /* insert all URIs into the array for which we have thumbnails now */
-  for (n = 0; uris[n] != NULL; n++)
-    dbus_message_iter_append_basic (&strv_iter, DBUS_TYPE_STRING, &uris[n]);
-
-  /* finish the URI string array */
-  dbus_message_iter_close_container (&iter, &strv_iter);
+  dbus_message_iter_init_append (message, &iter);
+  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &info->handle);
 
   /* send the signal message over D-Bus */
-  dbus_connection_send (dbus_g_connection_get_connection (service->connection), 
+  dbus_connection_send (dbus_g_connection_get_connection (info->service->connection), 
                         message, NULL);
 
   /* free the allocated D-Bus message */
   dbus_message_unref (message);
 
+  scheduler_idle_info_free (info);
+
+  return FALSE;
 }
 
 
@@ -496,26 +644,20 @@ tumbler_service_scheduler_started (TumblerScheduler *scheduler,
                                    const gchar      *origin,
                                    TumblerService   *service)
 {
-  DBusMessageIter iter;
-  DBusMessage    *message;
+  SchedulerIdleInfo *info;
+  
+  g_return_if_fail (TUMBLER_IS_SCHEDULER (scheduler));
+  g_return_if_fail (origin != NULL && *origin != '\0');
+  g_return_if_fail (TUMBLER_IS_SERVICE (service));
 
-  /* create a D-Bus message for the started signal */
-  message = dbus_message_new_signal (THUMBNAILER_PATH, THUMBNAILER_IFACE, "Started");
+  info = g_slice_new0 (SchedulerIdleInfo);
 
-  /* define the destination (the thumbnailer client) if possible */
-  if (origin)
-    dbus_message_set_destination (message, origin);
+  info->scheduler = g_object_ref (scheduler);
+  info->handle = handle;
+  info->origin = g_strdup (origin);
+  info->service = g_object_ref (service);
 
-  /* append the request handle */
-  dbus_message_iter_init_append (message, &iter);
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_UINT32, &handle);
-
-  /* send the signal message over D-Bus */
-  dbus_connection_send (dbus_g_connection_get_connection (service->connection), 
-                        message, NULL);
-
-  /* free the allocated D-Bus message */
-  dbus_message_unref (message);
+  g_idle_add (tumbler_service_started_idle, info);
 }
 
 
@@ -542,6 +684,24 @@ tumbler_service_pre_unmount (TumblerService *service,
 
 
 
+static void
+scheduler_idle_info_free (SchedulerIdleInfo *info)
+{
+  if (info == NULL)
+    return;
+
+  g_free (info->message);
+  g_free (info->origin);
+  g_strfreev (info->uris);
+
+  g_object_unref (info->scheduler);
+  g_object_unref (info->service);
+
+  g_slice_free (SchedulerIdleInfo, info);
+}
+
+
+
 TumblerService *
 tumbler_service_new (DBusGConnection *connection,
                      TumblerRegistry *registry)
